// very simple syntax highlighting rules for "mcode" in ACE editor
// and code folding/unfolding

ace.define("ace/mode/mcode_highlight_rules",["require","exports","module","ace/lib/oop","ace/mode/text_highlight_rules"], function(require, exports, module){"use strict";
var oop = require("../lib/oop");
var TextHighlightRules = require("./text_highlight_rules").TextHighlightRules;
var McodeHighlightRules = function (options) {
    this.$rules = {
        "start": [
            {
                token: "comment",   // URL or email
                regex: /(https?:\S*)|(\w+@.\S*)/
            }, {
                token: "comment",
                regex: /\/\/\/|\/\/|⍝|#|\/\*|\*\//
            }, {
                token: "string",    // simple string
                regex: /"""|(['"`][^'"`]*['"`])/
            }, {
                token: "runtime",   // runtime calls
                regex: /(mcode\.\w+)|(_\.\w+)/
            }, {
                token: "error",     // transpiler or runtime errors
                regex: /error:|\w*Error:|failed:/
            }, {
                token: "constant",  // numbers, language keywords
                regex: /(\b\d+\.?\d*e?-?\b)|(_\.\w+)|(\w+‗)|undefined|null|NaN/
            }, {
                token: "keyword",   // statements
                regex: /∇|⍔|:|→|⋄|⌻|⌸|⍐|⍗|⌺|⍝|;|←|⍂|⍢|⍫/,
            }, {
                token: "support",   // built-ins and constants
                regex: /⍺|⍵|δ|\+|\-|\*|\/|=|\%|⥊|\^|∨|~|<|≤|≥|>|\||\&|⍬|π|ė|‗|⊙|∆|#|\?\?|!=|!/,
            }, {
                token: "entity",    // primitives and vector functions
                regex: /⍂|⎕|⍎|⊢|⍃|≈|≠|⍙|θ|∊|⍷|⌷|⍴|⍳|↑|↓|¨|⌿|⍀|⍣|@|⍄|\?\=|\?|↑|↓|⍕|«|»|⊂|⊃|∩|∪|⍋|⍒|⌹|∘|≡|≢|⍪|⌈|⌊|⍟|√|⍛|⍉|⌽|×|÷|!|@|¦|⍠/,
            }, {
                token: "storage",   // delimiters and arguments
                regex: /\{|\(|\[|\}|\)|\]/,
            }
            // \⍺|\⍵|\δ
        ]
    };
};
oop.inherits(McodeHighlightRules, TextHighlightRules);
exports.McodeHighlightRules = McodeHighlightRules;

});

/*
// old mcode from python
ace.define("ace/mode/folding/mcode",["require","exports","module","ace/lib/oop","ace/mode/folding/fold_mode"], function(require, exports, module){"use strict";
var oop = require("../../lib/oop");
var BaseFoldMode = require("./fold_mode").FoldMode;
var FoldMode = exports.FoldMode = function (markers) {
    // this.foldingStartMarker = new RegExp("([\\[{])(?:\\s*)$|(" + markers + ")(?:\\s*)(?:#.*)?$");
    // this.foldingStartMarker = new RegExp("∇|⍢");
    this.foldingStartMarker = /(∇|⍢|⍝)|(\/\*)/;
    // this.foldingStopMarker = /^[\s\*]*(\*\/)/;
    this.foldingStopMarker = /(\*\/)/;
    console.log('FoldMode');
};
oop.inherits(FoldMode, BaseFoldMode);
(function () {
    this.getFoldWidgetRange = function (session, foldStyle, row) {
        let line = session.getLine(row);
        console.log(row+':'+line);
        let match = line.match(this.foldingStartMarker);
        console.log('match = '+JSON.stringify(match));
        if (match[1]) {
            // if (match[1])
            //     return this.openingBracketBlock(session, match[1], row, match.index);
            // if (match[2])
            //     return this.indentationBlock(session, row, match.index + match[2].length);
            return this.indentationBlock(session, row);
        }
        if (match[2]) {
            let i = match.index;
            let range = session.getCommentFoldRange(row, 0, 1);
//            let range = session.getCommentFoldRange(row, i + match[2].length, 1);
            console.log('range 2 = '+range);
            return range;
            // return session.getCommentFoldRange(row, i, 1);
        }
        if (match[3]) {
            let i = match.index;
            let range = session.getCommentFoldRange(row, i + 2, -1);
//            let range = session.getCommentFoldRange(row, i + match[2].length, 1);
            console.log('range 3 = '+range);
            return range;
            // return session.getCommentFoldRange(row, i, 1);
        }
        match = line.match(this.foldingStopMarker);
        if (match) {
            let i = match.index + match[0].length;
        //     // if (match[1])
        //     //     return this.closingBracketBlock(session, match[1], row, i);
            return session.getCommentFoldRange(row, i, -1);
        }
    };
}).call(FoldMode.prototype);

});
*/

// zv add:
// var seq = 0;
// let dl = s => { console.log(seq+' '+s); seq++; }
// let jstr = s => JSON.stringify(s)

// from cstyle
ace.define("ace/mode/folding/cstyle",["require","exports","module","ace/lib/oop","ace/range","ace/mode/folding/fold_mode"], function(require, exports, module){"use strict";
var oop = require("../../lib/oop");
var Range = require("../../range").Range;
var BaseFoldMode = require("./fold_mode").FoldMode;
var FoldMode = exports.FoldMode = function (commentRegex) {
    // dl('Mcode FoldMode from JS');
    if (commentRegex) {
        this.foldingStartMarker = new RegExp(this.foldingStartMarker.source.replace(/\|[^|]*?$/, "|" + commentRegex.start));
        this.foldingStopMarker = new RegExp(this.foldingStopMarker.source.replace(/\|[^|]*?$/, "|" + commentRegex.end));
    }
};
oop.inherits(FoldMode, BaseFoldMode);
(function () {
    // this.foldingStartMarker = /([\{\[\(])[^\}\]\)]*$|^\s*(\/\*)/;
    this.foldingStartMarker = /([\{\[\(])[^\}\]\)]*$|^\s*(\/\*)|(∇\.*\S*\s+\S+|⍢\s+\S*)/;
    this.foldingStopMarker = /^[^\[\{\(]*([\}\]\)])|^[\s\*]*(\*\/)/;
    // dl('this.foldingStartMarker = '+this.foldingStartMarker);
    // dl('this.foldingStopMarker = '+this.foldingStopMarker);
    this.singleLineBlockCommentRe = /^\s*(\/\*).*\*\/\s*$/;
    this.tripleStarBlockCommentRe = /^\s*(\/\*\*\*).*\*\/\s*$/;
    this.startRegionRe = /^\s*(\/\*|\/\/)#?region\b/;
    this._getFoldWidgetBase = this.getFoldWidget;
    this.getFoldWidget = function (session, foldStyle, row) {
        var line = session.getLine(row);
        if (this.singleLineBlockCommentRe.test(line)) {
            if (!this.startRegionRe.test(line) && !this.tripleStarBlockCommentRe.test(line))
                return "";
        }
        var fw = this._getFoldWidgetBase(session, foldStyle, row);
        if (!fw && this.startRegionRe.test(line))
            return "start"; // lineCommentRegionStart
        // dl('getFoldWidget: '+jstr(fw));
        return fw;
    };
    let mcodeCmtRange = (session,row,mi) => {
        // dl('  mcodeCmtRange');
        let e = session.getLines(row,99999), n = 0, b;
        // dl('  e = '+jstr(e));
        for (b of e) {
            // dl(n+' b = '+b);
            if (/\*\//.test(b)) break;
            n++
        }
        return new Range( row, mi, n+row, 0 );
    }
    this.getFoldWidgetRange = function (session, foldStyle, row, forceMultiline) {
        var line = session.getLine(row);
        // dl('getFoldWidgetRange '+jstr({foldStyle:foldStyle,line:line}));
        // dl(' checking: '+line);
        if (this.startRegionRe.test(line))
            return this.getCommentRegionBlock(session, line, row);
        var match = line.match(this.foldingStartMarker);
        // dl('  start match = '+jstr(match));
        if (match) {
            var i = match.index;
            if (match[1])
                return this.openingBracketBlock(session, match[1], row, i);

            // zv add:
            if (match[3])
                return this.indentationBlock(session, row, match.index + match[3].length);

            var range = session.getCommentFoldRange(row, i + match[0].length, 1);
            // dl('getCommentFoldRange range = '+jstr(range));

            let mi = i + match[0].length;
            range = mcodeCmtRange(session,row,mi);

            if (range && !range.isMultiLine()) {
                if (forceMultiline) {
                    range = this.getSectionRange(session, row);
                }
                else if (foldStyle != "all")
                    range = null;
            }
            return range;
        }
        if (foldStyle === "markbegin")
            return;
        var match = line.match(this.foldingStopMarker);
        // dl('  stop match = '+jstr(match));
        if (match) {
            var i = match.index + match[0].length;
            if (match[1])
                return this.closingBracketBlock(session, match[1], row, i);
            return session.getCommentFoldRange(row, i, -1);
        }
    };
    this.getSectionRange = function (session, row) {
        var line = session.getLine(row);
        var startIndent = line.search(/\S/);
        var startRow = row;
        var startColumn = line.length;
        row = row + 1;
        var endRow = row;
        var maxRow = session.getLength();
        while (++row < maxRow) {
            line = session.getLine(row);
            var indent = line.search(/\S/);
            if (indent === -1)
                continue;
            if (startIndent > indent)
                break;
            var subRange = this.getFoldWidgetRange(session, "all", row);
            if (subRange) {
                if (subRange.start.row <= startRow) {
                    break;
                }
                else if (subRange.isMultiLine()) {
                    row = subRange.end.row;
                }
                else if (startIndent == indent) {
                    break;
                }
            }
            endRow = row;
        }
        return new Range(startRow, startColumn, endRow, session.getLine(endRow).length);
    };
    this.getCommentRegionBlock = function (session, line, row) {
        var startColumn = line.search(/\s*$/);
        var maxRow = session.getLength();
        var startRow = row;
        var re = /^\s*(?:\/\*|\/\/|--)#?(end)?region\b/;
        var depth = 1;
        while (++row < maxRow) {
            line = session.getLine(row);
            var m = re.exec(line);
            if (!m)
                continue;
            if (m[1])
                depth--;
            else
                depth++;
            if (!depth)
                break;
        }
        var endRow = row;
        if (endRow > startRow) {
            return new Range(startRow, startColumn, endRow, line.length);
        }
    };
}).call(FoldMode.prototype);

});


ace.define("ace/mode/mcode",["require","exports","module","ace/lib/oop","ace/mode/text","ace/mode/mcode_highlight_rules","ace/mode/folding/cstyle"], function(require, exports, module){"use strict";
var oop = require("../lib/oop");
var TextMode = require("./text").Mode;
var McodeHighlightRules = require("./mcode_highlight_rules").McodeHighlightRules;
// var McodeFoldMode = require("./folding/mcode").FoldMode;
// var Range = require("../range").Range;
var CStyleFoldMode = require("./folding/cstyle").FoldMode;
var Mode = function () {
    this.HighlightRules = McodeHighlightRules;
    // this.$behaviour = this.$defaultBehaviour;
//    this.foldingRules = new McodeFoldMode("\\:");
    this.foldingRules = new CStyleFoldMode();
};
oop.inherits(Mode, TextMode);

(function () {
    this.lineCommentStart = "//";
    this.blockComment = { start: "/*", end: "*/" };
    this.$quotes = { '"': '"', "'": "'", "`": "`" };
    this.$pairQuotesAfter = {
        "`": /\w/
    };
    this.$id = "ace/mode/mcode";
    this.snippetFileId = "ace/snippets/mcode";
}).call(Mode.prototype);
exports.Mode = Mode;

});                (function() {
                    ace.require(["ace/mode/mcode"], function(m) {
                        if (typeof module == "object" && typeof exports == "object" && module) {
                            module.exports = m;
                        }
                    });
                })();
